import ast
import os
from typing import Dict, List, Optional, Callable, Tuple

import yaml
from graphql import GraphQLSchema, parse, build_ast_schema, build_client_schema, print_ast, print_schema
from pydantic import AnyHttpUrl, ValidationError
from rich import get_console

from turms.config import (
    GeneratorConfig,
    GraphQLConfigMultiple,
    GraphQLConfigSingle,
    GraphQLProject,
    SchemaType,
    LogFunction,
)
from turms.helpers import (
    load_introspection_from_glob,
    load_introspection_from_url,
    load_dsl_from_glob,
    load_dsl_from_url,
    import_string,
)
from turms.plugins.base import Plugin
from turms.parsers.base import Parser
from turms.processors.base import Processor
from turms.registry import ClassRegistry
from turms.stylers.base import Styler
from pydantic import ValidationError

from .errors import GenerationError
import json
import os


try:
    # If toml is installed, use it to load the config file
    import toml

    def toml_loader(file):
        return toml.loads(file.read())

except ImportError:

    def toml_loader(file):
        raise NotImplementedError("TOML not supported. Please install `toml`")


def json_loader(file):
    return json.loads(file.read())


SCANNABLE_FILE_NAMES = [
    "graphql.config.yaml",
    ".graphqlrc.yaml",
    "graphql.config.yml",
    ".graphqlrc.yml",
    "graphql.config.toml",
    ".graphqlrc.toml",
    "graphql.config.json",
    ".graphqlrc.json",
]

FILE_NAME_LOADERS = {
    "yaml": yaml.safe_load,
    "yml": yaml.safe_load,
    "toml": toml_loader,
    "json": json_loader,
}


def get_file_loader(file_name: str) -> Callable:
    try:
        return FILE_NAME_LOADERS[file_name.lower().split(".")[-1]]
    except AttributeError as err:
        raise GenerationError(
            f"File {file_name} is not supported. Please use one of {FILE_NAME_LOADERS.keys()}"
        ) from err


def load_projects_from_configpath(
    config_path: str, select: str = None
) -> Dict[str, GraphQLProject]:
    """Loads the configuration from a configuration file

    Args:
        config_path (str): The path to the config file

    Returns:
        GraphQLConfig: The configuration
    """
    file_path, file_name = os.path.split(config_path)

    with open(config_path, "r", encoding="utf-8") as file:
        loaded_dict = get_file_loader(config_path)(file)

    try:
        if "projects" in loaded_dict:
            projects = GraphQLConfigMultiple(**loaded_dict).projects
        else:
            projects = {"default": GraphQLConfigSingle(**loaded_dict)}
    except ValidationError as err:
        raise GenerationError(
            f"File {file_name} at {file_path} does not conform with turms."
        ) from err

    if select:
        projects = {key: project for key, project in projects.items() if key == select}
        assert len(projects) >= 1, "At least one project must be selected"

    return projects


def scan_folder_for_configs(folder_path: str = None) -> List[str]:
    """Scans a folder for config files

    Args:
        folder_path (str): The path to the folder

    Returns:
        List[str]: The list of config files
    """
    if folder_path is None:
        folder_path = os.getcwd()

    return [
        os.path.join(folder_path, file_name)
        for file_name in os.listdir(folder_path)
        if file_name.lower() in SCANNABLE_FILE_NAMES
    ]


def scan_folder_for_single_config(folder_path: str = None) -> str:
    """Scans a folder for one single config file

    Args:
        folder_path (str): The path to the folder

    Returns:
        str: The config file
    """

    configs = scan_folder_for_configs(folder_path=folder_path)

    if len(configs) == 0:
        raise GenerationError(
            f"No config files found in {folder_path}. Please use one of {SCANNABLE_FILE_NAMES}. Or use the --config flag to specify a config file."
        )

    if len(configs) != 1:
        raise GenerationError(
            f"Multiple config files found in {folder_path}. Please only have one of {SCANNABLE_FILE_NAMES}. Or use the --config flag to specify a config file."
        )

    return configs[0]


def write_code_to_file(code: str, outdir: str, filepath: str):
    if not os.path.isdir(outdir):  # pragma: no cover
        os.makedirs(outdir)

    generated_file = os.path.join(
        outdir,
        filepath,
    )

    with open(
        generated_file,
        "w",
        encoding="utf-8",
    ) as file:
        file.write(code)

    return generated_file


def write_schema_to_file(schema: GraphQLSchema, outdir: str, filepath: str):
    if not os.path.isdir(outdir):  # pragma: no cover
        os.makedirs(outdir)

    generated_file = os.path.join(
        outdir,
        filepath,
    )

    with open(
        generated_file,
        "w",
        encoding="utf-8",
    ) as file:
        file.write(print_schema(schema))

    return generated_file


def write_project(project: GraphQLProject, outdir: str, filepath: str):
    if not os.path.isdir(outdir):  # pragma: no cover
        os.makedirs(outdir)

    generated_file = os.path.join(
        outdir,
        filepath,
    )

    with open(
        generated_file,
        "w",
        encoding="utf-8",
    ) as file:
        file.write(project.model_dump_json(indent=4))

    return generated_file


def gen(
    filepath: Optional[str] = None,
    project_name: Optional[str] = None,
    strict: bool = False,
    overwrite_path: Optional[str] = None,
):
    """Generates  Code according to the config file

    Args:
        filepath (str, optional): The filepath of  graphqlconfig. Defaults to "graphql.config.yaml".
        project (str, optional): The project within that should be generated. Defaults to None.
    """

    if filepath is None:
        filepath = scan_folder_for_single_config()

    projects = load_projects_from_configpath(filepath, project_name)

    for key, project in projects.items():
        try:
            get_console().print(
                f"-------------- Generating project: {key} --------------"
            )

            generated_code = generate(project)

            write_code_to_file(
                generated_code,
                overwrite_path or project.extensions.turms.out_dir,
                project.extensions.turms.generated_name,
            )

            if project.extensions.turms.dump_schema:
                schema = build_schema_from_schema_type(
                            project.schema_url,
                            allow_introspection=project.extensions.turms.allow_introspection,
                        )
                
                write_schema_to_file(schema, project.extensions.turms.out_dir, project.extensions.turms.schema_name)

            if project.extensions.turms.dump_configuration:
                write_project(project, project.extensions.turms.out_dir, project.extensions.turms.configuration_name)






            get_console().print("Sucessfull!! :right-facing_fist::left-facing_fist:")
        except Exception as e:
            get_console().print(
                "-------- ERROR FOR PROJECT: " + key.upper() + " --------"
            )
            get_console().print_exception()
            if strict:
                raise GenerationError from e


def instantiate(module_path: str, **kwargs):
    """Instantiate A class from a file.

    Needs to conform to `path.to.module.ClassName`

    Args:
        module_path (str): The class path you would like to instatiate

    Returns:
        object: The instatiated class.
    """
    return import_string(module_path)(**kwargs)


def is_url(url: str) -> bool:
    return url.startswith("http") or url.startswith("https")


def build_schema_from_schema_type(
    schema: SchemaType, allow_introspection: bool = False
) -> GraphQLSchema:
    """Builds a schema from a project

    Args:
        project (GraphQLProject): The project

    Returns:
        GraphQLSchema: The schema
    """
    if isinstance(schema, dict):
        if len(schema.values()) == 1:
            key, value = list(schema.items())[0]
            try:
                dsl_string = load_dsl_from_url(key, value.headers)
                return build_ast_schema(parse(dsl_string))
            except Exception as e:
                if allow_introspection:
                    intropection = load_introspection_from_url(key, value.headers)
                    return build_client_schema(intropection)
                raise e
        else:
            # Multiple schemas, now we only support dsl
            dsl_subschemas = []

            for key, value in schema.items():
                dsl_subschemas.append(load_dsl_from_url(key, value.headers))

            return build_ast_schema(parse(" ".join(dsl_subschemas)))

    if isinstance(schema, list):
        if len(schema) == 1:
            # Only one schema, probably because of aesthetic reasons
            return build_schema_from_schema_type(
                schema[0], allow_introspection=allow_introspection
            )

        else:
            dsl_subschemas = []

            for item in schema:
                if is_url(item):
                    dsl_subschemas.append(load_dsl_from_url(item))
                if isinstance(item, dict):
                    for key, value in item.items():
                        dsl_subschemas.append(load_dsl_from_url(key, value.headers))
                if isinstance(item, str):
                    dsl_subschemas.append(load_dsl_from_glob(item))

            return build_ast_schema(parse(" ".join(dsl_subschemas)))

    if is_url(schema):
        try:
            dsl_string = load_dsl_from_url(schema)
            return build_ast_schema(parse(dsl_string))
        except Exception as e:
            if allow_introspection:
                intropection = load_introspection_from_url(schema)
                return build_client_schema(intropection)
            raise e

    if isinstance(schema, str):
        try:
            dsl_string = load_dsl_from_glob(schema)
            return build_ast_schema(parse(dsl_string))
        except Exception as e:
            if allow_introspection:
                intropection = load_introspection_from_glob(schema)
                return build_client_schema(intropection)
            raise e

    raise GenerationError("Could not build schema with type " + str(type(schema)))


def generate(project: GraphQLProject, log: Optional[LogFunction] = None) -> Tuple[str, GraphQLSchema ]:
    """Genrates the code according to the configugration

    The code is generated in the following order:
        1. Introspect the schema (either url or locally)
        2. Generate the of grapqhl.ast from this schema
        3. Instantiate all plugins/parsers/stylers
        4. Generate the ast from the ast through the plugins and stylers
        5. Parse the Ast with the parsers
        6. Generate the code from the ast through ast.unparse
        7. Process the code string through the processors

    Args:
        project (GraphQLConfig): The configuraion for the generation

    Returns:
        str: The generated code
    """
    if not log:

        def log(x, **kwargs):
            return print(x)

    gen_config = project.extensions.turms

    schema = build_schema_from_schema_type(
        project.schema_url,
        allow_introspection=project.extensions.turms.allow_introspection,
    )

    gen_config.documents = gen_config.documents or project.documents
    verbose = gen_config.verbose

    plugins = []
    stylers = []
    parsers = []
    processors = []

    for parser_config in gen_config.parsers:
        styler = instantiate(
            parser_config.type, config=parser_config.model_dump(), log=log
        )
        if verbose:
            get_console().print(f"Using Parser {styler}")
        parsers.append(styler)

    for plugins_config in gen_config.plugins:
        styler = instantiate(
            plugins_config.type, config=plugins_config.model_dump(), log=log
        )
        if verbose:
            get_console().print(f"Using Plugin {styler}")
        plugins.append(styler)

    for styler_config in gen_config.stylers:
        styler = instantiate(
            styler_config.type, config=styler_config.model_dump(), log=log
        )
        if verbose:
            get_console().print(f"Using Styler {styler}")
        stylers.append(styler)

    for proc_config in gen_config.processors:
        styler = instantiate(proc_config.type, config=proc_config.model_dump(), log=log)
        if verbose:
            get_console().print(f"Using Processor {styler}")
        processors.append(styler)

    code = generate_code(
        gen_config,
        schema,
        plugins=plugins,
        stylers=stylers,
        parsers=parsers,
        processors=processors,
        log=log,
    )

    return code, schema


def parse_asts_to_string(generated_ast: List[ast.AST]) -> str:
    module = ast.Module(body=generated_ast, type_ignores=[])
    return ast.unparse(ast.fix_missing_locations(module))


def generate_ast(
    config: GeneratorConfig,
    schema: GraphQLSchema,
    plugins: Optional[List[Plugin]] = None,
    stylers: Optional[List[Styler]] = None,
    skip_forwards: bool = False,
    log: LogFunction = lambda *args, **kwargs: print,
) -> List[ast.AST]:
    """Generates the ast from the schema

    Args:
        config (GeneratorConfig): The generaion Config (turms section)
        schema (GraphQLSchema): The schema to generate the ast from
        plugins (List[Plugins], optional): The plugins to use. Defaults to [].
        stylers (List[Styler], optional): The plugins to use. Defaults to [].

    Raises:
        GenerationError: Errors involving the generation of the ast

    Returns:
        List[ast.AST]: The generated ast (as list, not as module)
    """

    plugins = plugins or []
    stylers = stylers or []

    global_tree = []
    registry = ClassRegistry(config, stylers, log)

    for plugin in plugins:
        try:
            global_tree += plugin.generate_ast(schema, config, registry)
        except Exception as e:
            raise GenerationError(
                f"{plugin.__class__.__name__} failed!\n {str(e)}"
            ) from e

    global_tree = (
        registry.generate_imports() + registry.generate_builtins() + global_tree
    )
    if not skip_forwards:
        global_tree += registry.generate_forward_refs()

    return global_tree


def parse_ast(
    config: GeneratorConfig,
    ast: List[ast.AST],
    parsers: Optional[List[Parser]] = None,
    log: LogFunction = lambda *args, **kwargs: print,
) -> List[ast.AST]:
    """Parses the ast with the plugins

    Args:
        ast (List[ast.AST]): The ast to parse
        plugins (Optional[List[Plugin]], optional): The plugins to use. Defaults to [].
        stylers (Optional[List[Styler]], optional): The plugins to use. Defaults to [].

    Raises:
        GenerationError: Errors involving the generation of the ast

    Returns:
        List[ast.AST]: The parsed ast (as list, not as module)
    """

    parsers = parsers or []

    for parser in parsers:
        try:
            ast = parser.parse_ast(ast)
        except Exception as e:
            raise GenerationError(
                f"{parser.__class__.__name__} failed!\n {str(e)}"
            ) from e

    return ast


def process_code(
    config: GeneratorConfig,
    code: List[ast.AST],
    processors: Optional[List[Processor]] = None,
    log: LogFunction = lambda *args, **kwargs: print,
) -> List[ast.AST]:
    """Parses the ast with the plugins

    Args:
        ast (List[ast.AST]): The ast to parse
        plugins (Optional[List[Plugin]], optional): The plugins to use. Defaults to [].
        stylers (Optional[List[Styler]], optional): The plugins to use. Defaults to [].

    Raises:
        GenerationError: Errors involving the generation of the ast

    Returns:
        List[ast.AST]: The parsed ast (as list, not as module)
    """

    processors = processors or []

    for processor in processors:
        try:
            code = processor.run(code, config)
        except Exception as e:
            raise GenerationError(
                f"{processor.__class__.__name__} failed!\n {str(e)}"
            ) from e

    return code


def generate_code(
    config: GeneratorConfig,
    schema: GraphQLSchema,
    plugins: Optional[List[Plugin]] = None,
    stylers: Optional[List[Styler]] = None,
    parsers: Optional[List[Parser]] = None,
    processors: Optional[List[Processor]] = None,
    log: LogFunction = lambda *args, **kwargs: print,
) -> str:
    generated_ast = generate_ast(
        config,
        schema,
        plugins=plugins,
        stylers=stylers,
        skip_forwards=config.skip_forwards,
        log=log,
    )

    parsed_ast = parse_ast(config, generated_ast, parsers=parsers, log=log)
    code = parse_asts_to_string(parsed_ast)

    processed_code = process_code(config, code, processors=processors, log=log)

    return processed_code



from typing import Optional, Type, TypeVar

from pydantic import BaseModel

T = TypeVar("T")


def query(model: Type[T], variables) -> T:
    return model(variables)  # pragma: nocover


async def aquery(model: Type[T], variables) -> T:
    return model(variables)  # pragma: nocover


def subscribe(model: Type[T], variables) -> T:
    yield model(variables)  # pragma: nocover
    yield model(variables)  # pragma: nocover


async def asubscribe(model: Type[T], variables) -> T:
    yield model(variables)  # pragma: nocover
    yield model(variables)  # pragma: nocover


class ExtraArguments(BaseModel):
    extra: Optional[str]


class ExtraOnOperations(BaseModel):
    extra: Optional[str]


class ExtraArg(BaseModel):
    extra: Optional[str]



from abc import abstractmethod

from pydantic import BaseModel, Field
from pydantic_settings import BaseSettings, SettingsConfigDict


class StylerConfig(BaseSettings):
    model_config = SettingsConfigDict(extra="forbid")
    type: str


class Styler(BaseModel):
    """Base class for all stylers

    Stylers are used style the classnames function names of the generated python code.
    YOu can enforce specific code styles on the generated python code like (snake_case or camelCase)

    If you change the fieldname of a field in the GraphQL schema, the stylers will be used to
    style the fieldname in the generated python code and an alias will be added to the field.
    """

    config: StylerConfig

    @abstractmethod
    def style_subscription_name(self, name: str) -> str:
        raise NotImplementedError("Plugin must overrwrite this")  # pragma: no cover

    @abstractmethod
    def style_mutation_name(self, name: str) -> str:
        raise NotImplementedError("Plugin must overrwrite this")  # pragma: no cover

    @abstractmethod
    def style_query_name(self, name: str) -> str:
        raise NotImplementedError("Plugin must overrwrite this")  # pragma: no cover

    @abstractmethod
    def style_enum_name(self, name: str) -> str:
        raise NotImplementedError("Plugin must overrwrite this")  # pragma: no cover

    @abstractmethod
    def style_fragment_name(self, name: str) -> str:
        raise NotImplementedError("Plugin must overrwrite this")  # pragma: no cover

    @abstractmethod
    def style_node_name(self, name: str) -> str:
        raise NotImplementedError("Plugin must overrwrite this")  # pragma: no cover

    @abstractmethod
    def style_parameter_name(self, name: str) -> str:
        raise NotImplementedError("Plugin must overrwrite this")  # pragma: no cover

    @abstractmethod
    def style_input_name(self, name: str) -> str:
        raise NotImplementedError("Plugin must overrwrite this")  # pragma: no cover

    @abstractmethod
    def style_object_name(self, name: str) -> str:
        raise NotImplementedError("Plugin must overrwrite this")  # pragma: no cover


class BaseStyler(Styler):
    """A styler that uses no styling on the generated python code."""

    def style_query_name(self, name: str) -> str:
        return name

    def style_mutation_name(self, name: str) -> str:
        return name

    def style_enum_name(self, name: str) -> str:
        return name

    def style_fragment_name(self, name: str) -> str:
        return name

    def style_node_name(self, name: str) -> str:
        return name

    def style_input_name(self, name: str) -> str:
        return name

    def style_parameter_name(self, name: str) -> str:
        return name

    def style_subscription_name(self, name: str) -> str:
        return name

    def style_object_name(self, name: str) -> str:
        return name



import ast
import glob
import re
from typing import List, Optional, Set, Union

from graphql import (
    BooleanValueNode,
    FloatValueNode,
    FragmentDefinitionNode,
    GraphQLEnumType,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLOutputType,
    GraphQLScalarType,
    GraphQLUnionType,
    IntValueNode,
    ListTypeNode,
    NamedTypeNode,
    NonNullTypeNode,
    NullValueNode,
    OperationDefinitionNode,
    SelectionSetNode,
    StringValueNode,
    ValueNode,
    parse,
    print_ast,
    validate,
)
from graphql.error.graphql_error import GraphQLError
from graphql.language.ast import DocumentNode, FieldNode, NameNode
from graphql.utilities.build_client_schema import GraphQLSchema

from turms.config import GeneratorConfig
from turms.errors import (
    GenerationError,
    NoEnumFound,
    NoInputTypeFound,
    NoScalarFound,
)
from turms.registry import ClassRegistry

from .config import GraphQLTypes

commentline_regex = re.compile(r"^.*#(.*)")


class FragmentNotFoundError(GenerationError):
    pass


class NoDocumentsFoundError(GenerationError):
    pass


class InvalidDocuments(GenerationError):
    pass


class NoScalarEquivalentDefined(GenerationError):
    pass


def target_from_node(node: FieldNode) -> str:
    """Extacts the field name from a FieldNode. If alias is present, it will be used instead of the name"""
    return (
        node.alias.value if hasattr(node, "alias") and node.alias else node.name.value
    )


def non_typename_fields(node: FieldNode) -> List[FieldNode]:
    """Returns all fields in a FieldNode that are not __typename"""
    if not node.selection_set:
        return []
    return [field for field in node.selection_set.selections if not (isinstance(field, FieldNode) and field.name.value == "__typename")]


def inspect_operation_for_documentation(operation: OperationDefinitionNode):
    """Checks for operation level documentatoin"""

    definition = operation.loc.source.body.splitlines()[
        operation.loc.source.get_location(operation.loc.start).line
        - 1 : operation.loc.source.get_location(
            operation.selection_set.selections[0].loc.start
        ).line
        - 1
    ]
    doc = []
    for line in definition:
        if line and line != "":
            x = commentline_regex.match(line)
            if x:
                doc.append(x.group(1))

    return "\n".join(doc) if doc else None


def generate_typename_field(
    typename: str, registry: ClassRegistry, config: GeneratorConfig
):
    """Generates the typename field a specific type, this will be used to determine the type of the object in the response"""

    registry.register_import("pydantic.Field")
    registry.register_import("typing.Optional")
    registry.register_import("typing.Literal")

    keywords = [
        ast.keyword(arg="alias", value=ast.Constant(value="__typename")),
        ast.keyword(arg="default", value=ast.Constant(value=typename)),
    ]
    if config.exclude_typenames:
        keywords.append(ast.keyword(arg="exclude", value=ast.Constant(value=True)))

    return ast.AnnAssign(
        target=ast.Name(id="typename", ctx=ast.Store()),
        annotation=ast.Subscript(
                value=ast.Name("Literal", ctx=ast.Load()),
                slice=ast.Constant(value=typename),
                ctx=ast.Load(),
            ),
        value=ast.Call(
            func=ast.Name(id="Field", ctx=ast.Load()),
            args=[],
            keywords=keywords,
        ),
        simple=1,
    )

def generate_generic_typename_field(
    registry: ClassRegistry, config: GeneratorConfig
):
    """Generates the typename field a specific type, this will be used to determine the type of the object in the response"""

    registry.register_import("pydantic.Field")
    registry.register_import("typing.Optional")
    registry.register_import("typing.Literal")

    keywords = [
        ast.keyword(arg="alias", value=ast.Constant(value="__typename")),
    ]
    if config.exclude_typenames:
        keywords.append(ast.keyword(arg="exclude", value=ast.Constant(value=True)))

    return ast.AnnAssign(
        target=ast.Name(id="typename", ctx=ast.Store()),
        annotation=ast.Name("str", ctx=ast.Load()),
        value=ast.Call(
            func=ast.Name(id="Field", ctx=ast.Load()),
            args=[],
            keywords=keywords,
        ),
        simple=1,
    )


def generate_config_dict(
    graphQLType: GraphQLTypes,
    config: GeneratorConfig,
    registy: ClassRegistry,
    typename: str = None,
):
    """Generates the config class for a specific type version 2

    It will append the config class to the registry, and set the frozen
    attribute for the class to True, if the freeze config is enabled and
    the type appears in the freeze list.

    It will also add config attributes to the class, if the type appears in
    'additional_config' in the config file.

    """

    config_keywords = []

    if config.freeze.enabled:
        if graphQLType in config.freeze.types:
            if config.freeze.exclude and typename in config.freeze.exclude:
                pass
            elif config.freeze.include and typename not in config.freeze.include:
                pass
            else:
                config_keywords.append(
                    ast.keyword(arg="frozen", value=ast.Constant(value=True))
                )

    if config.options.enabled:
        if graphQLType in config.options.types:
            if config.options.exclude and typename in config.options.exclude:
                pass
            elif config.options.include and typename not in config.options.include:
                pass
            else:
                if config.options.allow_mutation is not None:
                    config_keywords.append(
                        ast.keyword(
                            arg="allow_mutation",
                            value=ast.Constant(value=config.options.allow_mutation),
                        )
                    )

                if config.options.extra is not None:
                    config_keywords.append(
                        ast.keyword(
                            arg="extra", value=ast.Constant(value=config.options.extra)
                        )
                    )

                if config.options.validate_assignment is not None:
                    config_keywords.append(
                        ast.keyword(
                            arg="validate_assignment",
                            value=ast.Constant(
                                value=config.options.validate_assignment
                            ),
                        )
                    )

                if config.options.allow_population_by_field_name is not None:
                    config_keywords.append(
                        ast.keyword(
                            arg="populate_by_name",
                            value=ast.Constant(
                                value=config.options.allow_population_by_field_name
                            ),
                        )
                    )

                if config.options.orm_mode is not None:
                    config_keywords.append(
                        ast.keyword(
                            arg="orm_mode",
                            value=ast.Constant(value=config.options.orm_mode),
                        )
                    )

                if config.options.use_enum_values is not None:
                    config_keywords.append(
                        ast.keyword(
                            arg="use_enum_values",
                            value=ast.Constant(value=config.options.use_enum_values),
                        )
                    )

    if typename:
        if typename in config.additional_config:
            for key, value in config.additional_config[typename].items():
                config_keywords.append(
                    ast.keyword(arg=key, value=ast.Constant(value=value))
                )

    if len(config_keywords) > 0:
        registy.register_import("pydantic.ConfigDict")
        return [
            ast.Assign(
                targets=[ast.Name(id="model_config", ctx=ast.Store())],
                value=ast.Call(
                    func=ast.Name(id="ConfigDict", ctx=ast.Load()),
                    args=[],
                    keywords=config_keywords,
                ),
            )
        ]
    else:
        return []


def generate_config_class_pydantic(
    graphQLType: GraphQLTypes, config: GeneratorConfig, typename: str = None
):
    """Generates the config class for a specific type

    It will append the config class to the registry, and set the frozen
    attribute for the class to True, if the freeze config is enabled and
    the type appears in the freeze list.

    It will also add config attributes to the class, if the type appears in
    'additional_config' in the config file.

    """

    config_fields = []

    if config.freeze.enabled:
        if graphQLType in config.freeze.types:
            if config.freeze.exclude and typename in config.freeze.exclude:
                pass
            elif config.freeze.include and typename not in config.freeze.include:
                pass
            else:
                config_fields.append(
                    ast.Assign(
                        targets=[ast.Name(id="frozen", ctx=ast.Store())],
                        value=ast.Constant(value=True),
                    )
                )

    if config.options.enabled:
        if graphQLType in config.options.types:
            if config.options.exclude and typename in config.options.exclude:
                pass
            elif config.options.include and typename not in config.options.include:
                pass
            else:
                if config.options.allow_mutation is not None:
                    config_fields.append(
                        ast.Assign(
                            targets=[ast.Name(id="allow_mutation", ctx=ast.Store())],
                            value=ast.Constant(value=config.options.allow_mutation),
                        )
                    )

                if config.options.extra is not None:
                    config_fields.append(
                        ast.Assign(
                            targets=[ast.Name(id="extra", ctx=ast.Store())],
                            value=ast.Constant(value=config.options.extra),
                        )
                    )

                if config.options.validate_assignment is not None:
                    config_fields.append(
                        ast.Assign(
                            targets=[
                                ast.Name(id="validate_assignment", ctx=ast.Store())
                            ],
                            value=ast.Constant(
                                value=config.options.validate_assignment
                            ),
                        )
                    )

                if config.options.allow_population_by_field_name is not None:
                    config_fields.append(
                        ast.Assign(
                            targets=[
                                ast.Name(
                                    id="allow_population_by_field_name", ctx=ast.Store()
                                )
                            ],
                            value=ast.Constant(
                                value=config.options.allow_population_by_field_name
                            ),
                        )
                    )

                if config.options.orm_mode is not None:
                    config_fields.append(
                        ast.Assign(
                            targets=[ast.Name(id="orm_mode", ctx=ast.Store())],
                            value=ast.Constant(value=config.options.orm_mode),
                        )
                    )

                if config.options.use_enum_values is not None:
                    config_fields.append(
                        ast.Assign(
                            targets=[ast.Name(id="use_enum_values", ctx=ast.Store())],
                            value=ast.Constant(value=config.options.use_enum_values),
                        )
                    )

    if typename:
        if typename in config.additional_config:
            for key, value in config.additional_config[typename].items():
                config_fields.append(
                    ast.Assign(
                        targets=[ast.Name(id=key, ctx=ast.Store())],
                        value=ast.Constant(value=value),
                    )
                )

    if len(config_fields) > 0:
        config_fields.insert(
            0,
            ast.Expr(
                value=ast.Constant(value="A config class"),
            ),
        )
    if len(config_fields) > 0:
        return [
            ast.ClassDef(
                name="Config",
                bases=[],
                keywords=[],
                body=config_fields,
                decorator_list=[],
            )
        ]
    else:
        return []


def generate_pydantic_config(
    graphQLType: GraphQLTypes,
    config: GeneratorConfig,
    registry: ClassRegistry,
    typename: str = None,
):
    if config.pydantic_version == "v2":
        return generate_config_dict(graphQLType, config, registry, typename)
    else:
        return generate_config_class_pydantic(graphQLType, config, typename)



def add_typename_recursively(selection_set: SelectionSetNode, skip=False) -> None:
    if selection_set is None:
        return

    # Collect all existing fields in the selection set
    selections = list(selection_set.selections)
    has_typename = any(
        isinstance(field, FieldNode) and field.name.value == "__typename"
        for field in selections
    )

    # Add __typename if it's not already present
    if not has_typename and not skip:
        selections.append(
            FieldNode(
                name=NameNode(value="__typename"),
                arguments=[],
                directives=[],
                selection_set=None,
            )
        )

    # Apply the function recursively to nested selection sets
    for field in selections:
        if isinstance(field, FieldNode) and field.selection_set:
            add_typename_recursively(field.selection_set)

    # Update the selection set with potentially added __typename fields
    selection_set.selections = tuple(selections)

def auto_add_typename_field_to_all_objects(document: DocumentNode) -> DocumentNode:
    for definition in document.definitions:
        if isinstance(definition, (OperationDefinitionNode, FragmentDefinitionNode)):
            add_typename_recursively(definition.selection_set, skip=isinstance(definition, OperationDefinitionNode))
    
    
    return document


def parse_documents(client_schema: GraphQLSchema, scan_glob) -> DocumentNode:
    """ """
    if not scan_glob:
        raise GenerationError("Couldnt find documents glob")

    x = glob.glob(scan_glob, recursive=True)

    errors: List[GraphQLError] = []

    dsl = ""

    for file in x:
        with open(file, "r") as f:
            dsl += f.read()

    if dsl == "":
        raise NoDocumentsFoundError(
            f"Glob {scan_glob} did not find documents. Or only empty documents"
        )

    nodes = parse(dsl)

    errors = validate(client_schema, nodes)
    if len(errors) > 0:
        raise InvalidDocuments(
            "Invalid Documents \n" + "\n".join(str(e) for e in errors)
        )
    
    
    nodes = auto_add_typename_field_to_all_objects(nodes)

   

    return nodes


fragment_searcher = re.compile(r"\.\.\.(?P<fragment>[a-zA-Z]*)")



def auto_add_typename_field_to_fragment_str(fragment_str: str) -> str:
    x = parse(fragment_str)
    for fragment in x.definitions:
        if isinstance(fragment, FragmentDefinitionNode):
            selections = list(fragment.selection_set.selections)
            if not any(isinstance(field, FieldNode) and field.name.value == "__typename" for field in selections):
                selections.append(
                    FieldNode(
                        name=NameNode(value="__typename"),
                        arguments=[],
                        directives=[],
                        selection_set=None,
                    )
                )
                fragment.selection_set.selections = tuple(selections)





    return print_ast(x)






def replace_iteratively(
    pattern,
    registry,
    taken=[],
):
    z = set(fragment_searcher.findall(pattern))  # only set is important
    new_fragments = [new_f for new_f in z if new_f not in taken and new_f != ""]
    if not new_fragments:
        return pattern
    else:
        try:
            level_down_pattern = "\n\n".join(
                [auto_add_typename_field_to_fragment_str(registry.get_fragment_document(key)) for key in new_fragments]
                + [pattern]
            )
            return replace_iteratively(
                level_down_pattern, registry, taken=new_fragments + taken
            )
        except KeyError as e:
            raise FragmentNotFoundError(
                f"Could not find in Fragment Map {registry}"
            ) from e


def get_additional_bases_for_type(
    typename, config: GeneratorConfig, registry: ClassRegistry
):
    if typename in config.additional_bases:
        for base in config.additional_bases[typename]:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in config.additional_bases[typename]
        ]
    return []


def get_interface_bases(config: GeneratorConfig, registry: ClassRegistry):
    if config.interface_bases:
        for base in config.interface_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in config.interface_bases
        ]
    else:
        for base in config.object_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in config.object_bases
        ]


def interface_is_extended_by_other_interfaces(
    interface: GraphQLInterfaceType, other_interfaces: Set[GraphQLInterfaceType]
) -> bool:
    interfaces_implemented_by_other_interfaces = {
        nested_interface
        for other_interface in other_interfaces
        for nested_interface in other_interface.interfaces
    }
    return interface in interfaces_implemented_by_other_interfaces


def recurse_type_annotation(
    type: NamedTypeNode,
    registry: ClassRegistry,
    optional=True,
    overwrite_final: Optional[str] = None,
):
    if isinstance(type, NonNullTypeNode):
        return recurse_type_annotation(
            type.type, registry, optional=False, overwrite_final=overwrite_final
        )

    if isinstance(type, ListTypeNode):
        if optional:
            registry.register_import("typing.List")
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name(id="Optional", ctx=ast.Load()),
                slice=ast.Subscript(
                    value=ast.Name(id="List", ctx=ast.Load()),
                    slice=recurse_type_annotation(
                        type.type, registry, overwrite_final=overwrite_final
                    ),
                ),
            )

        registry.register_import("typing.List")
        return ast.Subscript(
            value=ast.Name(id="List", ctx=ast.Load()),
            slice=recurse_type_annotation(
                type.type, registry, overwrite_final=overwrite_final
            ),
        )

    if isinstance(type, NamedTypeNode):
        x = None
        if overwrite_final is not None:
            x = ast.Name(id=overwrite_final, ctx=ast.Load())
        else:
            try:
                x = registry.reference_scalar(type.name.value)
            except NoScalarFound:
                try:
                    x = registry.reference_inputtype(
                        type.name.value, "", allow_forward=False
                    )
                except NoInputTypeFound:
                    try:
                        x = registry.reference_enum(
                            type.name.value, "", allow_forward=False
                        )
                    except NoEnumFound:
                        raise NotImplementedError(
                            f"Could not find corresponding type for type '{type.name.value}'. Did you register this scalar?"
                        )

        if not x:
            raise Exception(f"Could not set value for {type}")

        if optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name(id="Optional", ctx=ast.Load()),
                slice=x,
            )

        return x

    raise NotImplementedError("oisnosin")


def recurse_outputtype_annotation(
    type: GraphQLOutputType,
    registry: ClassRegistry,
    optional=True,
    overwrite_final: Optional[str] = None,
) -> ast.expr:
    if isinstance(type, GraphQLNonNull):
        return recurse_outputtype_annotation(
            type.of_type, registry, optional=False, overwrite_final=overwrite_final
        )

    if isinstance(type, GraphQLList):
        if optional:
            registry.register_import("typing.List")
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name(id="Optional", ctx=ast.Load()),
                slice=ast.Subscript(
                    value=ast.Name(id="List", ctx=ast.Load()),
                    slice=recurse_outputtype_annotation(
                        type.of_type, registry, overwrite_final=overwrite_final
                    ),
                ),
            )

        registry.register_import("typing.List")
        return ast.Subscript(
            value=ast.Name(id="List", ctx=ast.Load()),
            slice=recurse_outputtype_annotation(
                type.of_type, registry, overwrite_final=overwrite_final
            ),
        )

    if isinstance(type, GraphQLEnumType):
        if optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name(id="Optional", ctx=ast.Load()),
                slice=registry.reference_enum(type.name, "", allow_forward=False),
            )

        return registry.reference_enum(type.name, "", allow_forward=False)

    if isinstance(type, GraphQLScalarType):
        if optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_scalar(type.name),
            )

        else:
            return registry.reference_scalar(type.name)

    if isinstance(type, (GraphQLObjectType, GraphQLInterfaceType, GraphQLUnionType)):
        assert overwrite_final, "Needs to be set"
        if optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=ast.Name(id=overwrite_final, ctx=ast.Load()),
            )

        else:
            return ast.Name(id=overwrite_final, ctx=ast.Load())

    raise NotImplementedError(f"recurse over {type.__class__.__name__}")  # pragma: no cover


def recurse_outputtype_label(
    type: GraphQLOutputType,
    registry: ClassRegistry,
    optional=True,
    overwrite_final: Optional[str] = None,
):
    if isinstance(type, GraphQLNonNull):  # pragma: no cover
        return recurse_outputtype_label(
            type.of_type, registry, optional=False, overwrite_final=overwrite_final
        )

    if isinstance(type, GraphQLList):
        if optional:
            return (
                "Optional[List["
                + recurse_outputtype_label(
                    type.of_type, registry, overwrite_final=overwrite_final
                )
                + "]]"
            )

        return (
            "List["
            + recurse_outputtype_label(
                type.of_type, registry, overwrite_final=overwrite_final
            )
            + "]"
        )

    if isinstance(type, GraphQLEnumType):
        if optional:
            return (
                "Optional["
                + registry.reference_enum(type.name, "", allow_forward=False).id
                + "]"
            )

        return registry.reference_enum(type.name, "", allow_forward=False).id

    if isinstance(type, GraphQLScalarType):
        if optional:
            return "Optional[" + registry.reference_scalar(type.name).id + "]"

        else:
            return registry.reference_scalar(type.name).id

    if isinstance(type, (GraphQLObjectType, GraphQLInterfaceType, GraphQLUnionType)):
        assert overwrite_final, "Needs to be set"
        if optional:
            return "Optional[" + overwrite_final + "]"

        else:
            return overwrite_final

    raise NotImplementedError("oisnosin")


def recurse_type_label(
    type: NamedTypeNode,
    registry: ClassRegistry,
    optional=True,
    overwrite_final: Optional[str] = None,
):
    if isinstance(type, NonNullTypeNode):
        return recurse_type_label(
            type.type, registry, optional=False, overwrite_final=overwrite_final
        )

    if isinstance(type, ListTypeNode):
        if optional:
            return (
                "Optional[List["
                + recurse_type_label(
                    type.type, registry, overwrite_final=overwrite_final
                )
                + "]]"
            )

        return (
            "List["
            + recurse_type_label(type.type, registry, overwrite_final=overwrite_final)
            + "]"
        )

    if isinstance(type, NamedTypeNode):
        if overwrite_final is not None:
            x = ast.Name(id=overwrite_final, ctx=ast.Load())
        else:
            try:
                x = registry.reference_scalar(type.name.value)
            except NoScalarFound:
                try:
                    x = registry.reference_inputtype(
                        type.name.value, "", allow_forward=False
                    )
                except NoInputTypeFound:
                    try:
                        x = registry.reference_enum(
                            type.name.value, "", allow_forward=False
                        )
                    except NoEnumFound:
                        raise NotImplementedError(
                            f"Could not find correspoinding type labler for {type.name.value}"
                        )

        if optional:
            return "Optional[" + x.id + "]"

        return x.id


def parse_value_node(value_node: ValueNode) -> Union[None, str, int, float, bool]:
    """Parses a Value Node into a Python value
    using standard types

    Args:
        value_node (ValueNode): The Argument Value Node

    Raises:
        NotImplementedError: If the Value Node is not supported

    Returns:
        Union[None, str, int, float, bool]: The parsed value
    """
    if isinstance(value_node, IntValueNode):
        return int(value_node.value)
    elif isinstance(value_node, FloatValueNode):
        return float(value_node.value)
    elif isinstance(value_node, StringValueNode):
        return value_node.value
    elif isinstance(value_node, BooleanValueNode):
        return value_node.value == "true"
    elif isinstance(value_node, NullValueNode):
        return None
    else:
        raise NotImplementedError(f"Cannot parse {value_node}")



from graphql import (
    GraphQLField,
    GraphQLInputObjectType,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLType,
    GraphQLUnionType,
)
from pydantic_settings import SettingsConfigDict
from turms.errors import GenerationError
from turms.plugins.base import Plugin, PluginConfig
import ast
from typing import Dict, List
from turms.config import GeneratorConfig
from graphql.utilities.build_client_schema import GraphQLSchema
from pydantic import Field
from graphql.type.definition import (
    GraphQLEnumType,
)
from turms.registry import ClassRegistry
from turms.utils import (
    generate_pydantic_config,
    get_additional_bases_for_type,
    interface_is_extended_by_other_interfaces,
)
from turms.config import GraphQLTypes


class ObjectsPluginConfig(PluginConfig):
    model_config = SettingsConfigDict(env_prefix="TURMS_PLUGINS_OBJECTS_")
    type: str = "turms.plugins.objects.ObjectsPlugin"
    types_bases: List[str] = ["pydantic.BaseModel"]
    skip_underscore: bool = False
    skip_double_underscore: bool = True


def generate_object_field_annotation(
    graphql_type: GraphQLType,
    parent: str,
    config: GeneratorConfig,
    plugin_config: ObjectsPluginConfig,
    registry: ClassRegistry,
    is_optional=True,
):
    if isinstance(graphql_type, GraphQLScalarType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_scalar(graphql_type.name),
                ctx=ast.Load(),
            )

        return registry.reference_scalar(graphql_type.name)

    if isinstance(graphql_type, GraphQLInterfaceType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_interface(graphql_type.name, parent),
                ctx=ast.Load(),
            )

        return registry.reference_interface(graphql_type.name, parent)

    if isinstance(graphql_type, GraphQLObjectType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_object(graphql_type.name, parent),
                ctx=ast.Load(),
            )
        return registry.reference_object(graphql_type.name, parent)

    if isinstance(graphql_type, GraphQLUnionType):
        if is_optional:
            registry.register_import("typing.Optional")
            registry.register_import("typing.Union")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=ast.Subscript(
                    value=ast.Name("Union", ctx=ast.Load()),
                    slice=ast.Tuple(
                        elts=[
                            generate_object_field_annotation(
                                union_type,
                                parent,
                                config,
                                plugin_config,
                                registry,
                                is_optional=False,
                            )
                            for union_type in graphql_type.types
                        ],
                        ctx=ast.Load(),
                    ),
                ),
                ctx=ast.Load(),
            )
        registry.register_import("typing.Union")

        return ast.Subscript(
            value=ast.Name("Union", ctx=ast.Load()),
            slice=ast.Tuple(
                elts=[
                    generate_object_field_annotation(
                        union_type,
                        parent,
                        config,
                        plugin_config,
                        registry,
                        is_optional=False,
                    )
                    for union_type in graphql_type.types
                ],
                ctx=ast.Load(),
            ),
            ctx=ast.Load(),
        )

    if isinstance(graphql_type, GraphQLEnumType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_enum(
                    graphql_type.name,
                    parent,
                    allow_forward=not config.force_plugin_order,
                ),
                ctx=ast.Load(),
            )
        return registry.reference_enum(
            graphql_type.name, parent, allow_forward=not config.force_plugin_order
        )

    if isinstance(graphql_type, GraphQLNonNull):
        return generate_object_field_annotation(
            graphql_type.of_type,
            parent,
            config,
            plugin_config,
            registry,
            is_optional=False,
        )

    if isinstance(graphql_type, GraphQLList):
        if (
            config.freeze.enabled
            and GraphQLTypes.OBJECT in config.freeze.types
            and config.freeze.convert_list_to_tuple
        ):
            registry.register_import("typing.Tuple")

            def list_builder(x):
                return ast.Subscript(
                    value=ast.Name("Tuple", ctx=ast.Load()),
                    slice=ast.Tuple(elts=[x, ast.Constant(value=...)], ctx=ast.Load()),
                    ctx=ast.Load(),
                )

        else:

            registry.register_import("typing.List")

            def list_builder(x):
                return ast.Subscript(
                    value=ast.Name("List", ctx=ast.Load()), slice=x, ctx=ast.Load()
                )

        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=list_builder(
                    generate_object_field_annotation(
                        graphql_type.of_type,
                        parent,
                        config,
                        plugin_config,
                        registry,
                        is_optional=True,
                    )
                ),
                ctx=ast.Load(),
            )

        registry.register_import("typing.List")
        return list_builder(
            generate_object_field_annotation(
                graphql_type.of_type,
                parent,
                config,
                plugin_config,
                registry,
                is_optional=True,
            )
        )

    raise NotImplementedError(f"Unknown input type {repr(graphql_type)}")


def generate_types(
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: ObjectsPluginConfig,
    registry: ClassRegistry,
):

    tree = []

    objects = {
        key: value
        for key, value in client_schema.type_map.items()
        if isinstance(value, GraphQLObjectType)
        and not isinstance(value, GraphQLInputObjectType)
        or isinstance(value, GraphQLInterfaceType)
    }

    sorted_objects = {
        k: v
        for k, v in sorted(
            objects.items(),
            key=lambda item: isinstance(item[1], GraphQLInterfaceType),
            reverse=True,
        )
    }

    interface_map: Dict[str, List[str]] = (
        {}
    )  # A list of interfaces with the union classes attached
    interface_base_map: Dict[str, str] = (
        {}
    )  # A list of interfaces with its respective base

    for base in plugin_config.types_bases:
        registry.register_import(base)

    for key, object_type in sorted_objects.items():

        additional_bases = get_additional_bases_for_type(
            object_type.name, config, registry
        )

        if plugin_config.skip_underscore and key.startswith("_"):
            continue

        if plugin_config.skip_double_underscore and key.startswith("__"):
            continue

        if isinstance(object_type, GraphQLObjectType):
            classname = registry.generate_objecttype(key)
        if isinstance(object_type, GraphQLInterfaceType):
            classname = registry.generate_interface(key)
            interface_base_map[key] = classname

        for interface in object_type.interfaces:
            # Populate the Union_classed
            interface_map.setdefault(interface.name, []).append(classname)

            other_interfaces = set(object_type.interfaces) - {interface}
            if not interface_is_extended_by_other_interfaces(
                interface, other_interfaces
            ):
                additional_bases.append(
                    ast.Name(
                        id=registry.inherit_interface(interface.name),
                        ctx=ast.Load(),
                    )
                )

        fields = (
            [ast.Expr(value=ast.Constant(value=object_type.description))]
            if object_type.description
            else []
        )

        for value_key, value in object_type.fields.items():
            value: GraphQLField
            field_name = registry.generate_node_name(value_key)

            keywords = []
            if not isinstance(value.type, GraphQLNonNull):
                keywords.append(
                    ast.keyword(arg="default", value=ast.Constant(value=None))
                )

            if field_name != value_key:
                registry.register_import("pydantic.Field")
                assign = ast.AnnAssign(
                    target=ast.Name(field_name, ctx=ast.Store()),
                    annotation=generate_object_field_annotation(
                        value.type,
                        classname,
                        config,
                        plugin_config,
                        registry,
                        is_optional=True,
                    ),
                    value=ast.Call(
                        func=ast.Name(id="Field", ctx=ast.Load()),
                        args=[],
                        keywords=keywords
                        + [
                            ast.keyword(
                                arg="alias", value=ast.Constant(value=value_key)
                            )
                        ],
                    ),
                    simple=1,
                )
            else:
                if keywords:
                    registry.register_import("pydantic.Field")
                assign = ast.AnnAssign(
                    target=ast.Name(value_key, ctx=ast.Store()),
                    annotation=generate_object_field_annotation(
                        value.type,
                        classname,
                        config,
                        plugin_config,
                        registry,
                        is_optional=True,
                    ),
                    value=(
                        ast.Call(
                            func=ast.Name(id="Field", ctx=ast.Load()),
                            args=[],
                            keywords=keywords,
                        )
                        if keywords
                        else None
                    ),
                    simple=1,
                )

            potential_comment = (
                value.description
                if not value.deprecation_reason
                else f"DEPRECATED: {value.description}"
            )

            if potential_comment:
                fields += [
                    assign,
                    ast.Expr(value=ast.Constant(value=potential_comment)),
                ]

            else:
                fields += [assign]

        tree.append(
            ast.ClassDef(
                classname,
                bases=additional_bases
                + [
                    ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                    for base in plugin_config.types_bases
                ],
                decorator_list=[],
                keywords=[],
                body=fields
                + generate_pydantic_config(GraphQLTypes.OBJECT, config, registry, key),
            )
        )

    for interface, union_class_names in interface_map.items():
        registry.register_import("typing.Union")
        tree.append(
            ast.Assign(
                targets=[ast.Name(id=interface, ctx=ast.Store())],
                value=ast.Subscript(
                    value=ast.Name("Union", ctx=ast.Load()),
                    slice=ast.Tuple(
                        elts=[
                            ast.Name(id=clsname, ctx=ast.Load())
                            for clsname in union_class_names
                        ],
                        ctx=ast.Load(),
                    ),
                    ctx=ast.Load(),
                ),
            )
        )

    unimplemented_interfaces = {
        interface: baseclass
        for interface, baseclass in interface_base_map.items()
        if interface not in interface_map
    }
    if unimplemented_interfaces:
        if config.always_resolve_interfaces:
            raise GenerationError(
                f"Interfaces {unimplemented_interfaces.keys()} have no types implementing it. And we have set always_resolve_interfaces to true. Make sure your schema is correct"
            )

        for interface, baseclass in unimplemented_interfaces.items():
            registry.warn(f"Interface {interface} has no types implementing it")
            tree.append(
                ast.Assign(
                    targets=[ast.Name(id=interface, ctx=ast.Store())],
                    value=ast.Name(baseclass, ctx=ast.Load()),
                )
            )

    return tree


class ObjectsPlugin(Plugin):
    """Generate Pydantic models for GraphQL objects

    This plugin generates Pydantic models for GraphQL ObjectTypes and Interfaces in your graphql schema.

    Attention: This plugin is not made for client side usage. Most likely you are looking for the
    operations, fragments and funcs plugin (that generates client side validation and query function
    for your documents (your operations, fragments and queries).


    """

    config: ObjectsPluginConfig = Field(default_factory=ObjectsPluginConfig)

    def generate_ast(
        self,
        client_schema: GraphQLSchema,
        config: GeneratorConfig,
        registry: ClassRegistry,
    ) -> List[ast.AST]:

        return generate_types(client_schema, config, self.config, registry)



from pydantic import BaseModel, Field
from turms.stylers.base import BaseStyler


class CapitalizeSylerConfig(BaseModel):
    type: str = "turms.stylers.capitalize.CapitalizerSyler"
    capitalize_fragment: bool = True
    capitalize_query: bool = True
    capitalize_mutation: bool = True
    capitalize_subscription: bool = True
    capitalize_enum: bool = True
    capitalize_input: bool = True


class CapitalizeStyler(BaseStyler):
    """A styler that capitalizes the first letter of the python class names."""

    config: CapitalizeSylerConfig = Field(default_factory=CapitalizeSylerConfig)

    def style_fragment_name(self, typename):
        return typename[0].upper() + typename[1:]

    def style_query_name(self, typename):
        return typename[0].upper() + typename[1:]

    def style_subscription_name(self, typename):
        return typename[0].upper() + typename[1:]

    def style_mutation_name(self, typename):
        return typename[0].upper() + typename[1:]

    def style_input_name(self, typename):
        return typename[0].upper() + typename[1:]

    def style_enum_name(self, typename):
        return typename[0].upper() + typename[1:]



from abc import abstractmethod

from pydantic import BaseModel, Field
from pydantic_settings import BaseSettings, SettingsConfigDict
from turms.config import GeneratorConfig


class ProcessorConfig(BaseSettings):
    model_config = SettingsConfigDict(extra="forbid")
    type: str


class Processor(BaseModel):
    """Base class for all processors

    Processors are used to modify the generated python code before it is written to the file.
    You can use processors to enforce specific styles on the generated python code like (black
    or isort) or to add additional code to the generated python code.
    """

    config: ProcessorConfig

    @abstractmethod
    def run(gen_file: str, config: GeneratorConfig): ...  # pragma: no cover



from typing import Dict, Set

from graphql import (
    FragmentDefinitionNode,
    GraphQLEnumType,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLScalarType,
    ListTypeNode,
    NamedTypeNode,
    NonNullTypeNode,
    OperationDefinitionNode,
)
from graphql.language.ast import (
    DocumentNode,
    FieldNode,
    FragmentSpreadNode,
    InlineFragmentNode,
)
from graphql.type.definition import (
    GraphQLType,
    GraphQLUnionType,
)
from graphql.utilities.build_client_schema import GraphQLSchema


class ReferenceRegistry:
    def __init__(self):
        self.objects: Set[str] = set()
        self.fragments: Set[str] = set()
        self.enums: Set[str] = set()
        self.inputs: Set[str] = set()
        self.scalars: Set[str] = set()
        self.operations: Set[str] = set()

    def register_type(self, type_name: str):
        self.objects.add(type_name)

    def register_fragment(self, type_name: str):
        self.fragments.add(type_name)

    def register_enum(self, type_name: str):
        self.enums.add(type_name)

    def register_input(self, type_name: str):
        self.inputs.add(type_name)

    def register_scalar(self, type_name: str):
        self.scalars.add(type_name)

    def is_input_registered(self, type_name: str):
        return type_name in self.inputs


def recurse_find_references(
    node: FieldNode,
    graphql_type: GraphQLType,
    client_schema: GraphQLSchema,
    registry: ReferenceRegistry,
    is_optional=True,
):
    if isinstance(
        graphql_type, (GraphQLUnionType, GraphQLObjectType, GraphQLInterfaceType)
    ):
        for sub_node in node.selection_set.selections:
            if isinstance(sub_node, FragmentSpreadNode):
                registry.register_fragment(sub_node.name.value)

            elif isinstance(sub_node, InlineFragmentNode):
                for sub_sub_node in sub_node.selection_set.selections:
                    if isinstance(sub_sub_node, FieldNode):
                        sub_sub_node_type = client_schema.get_type(
                            sub_node.type_condition.name.value
                        )

                        if sub_sub_node.name.value == "__typename":
                            continue

                        field_type = sub_sub_node_type.fields[sub_sub_node.name.value]
                        recurse_find_references(
                            sub_sub_node,
                            field_type.type,
                            client_schema,
                            registry,
                        )
            elif isinstance(sub_node, FieldNode):
                if sub_node.name.value == "__typename":
                    continue
                this_type = graphql_type.fields[sub_node.name.value]

                recurse_find_references(
                    sub_node,
                    this_type.type,
                    client_schema,
                    registry,
                )
            else:
                raise Exception("Unexpected Type")

    elif isinstance(graphql_type, GraphQLScalarType):
        registry.register_scalar(graphql_type.name)

    elif isinstance(graphql_type, GraphQLEnumType):
        registry.register_enum(graphql_type.name)

    elif isinstance(graphql_type, GraphQLNonNull):
        recurse_find_references(
            node,
            graphql_type.of_type,
            client_schema,
            registry,
            is_optional=False,
        )

    elif isinstance(graphql_type, GraphQLList):
        recurse_find_references(
            node,
            graphql_type.of_type,
            client_schema,
            registry,
            is_optional=False,
        )
    else:
        raise Exception("Unknown Type", type(graphql_type), graphql_type)


def break_recursion_loop(*args, **kwargs):
    return recurse_type_annotation(*args, **kwargs)


def recurse_type_annotation(
    field: GraphQLInputField,
    graphql_type: NamedTypeNode,
    schema: GraphQLSchema,
    registry: ReferenceRegistry,
    optional=True,
):

    if isinstance(graphql_type, NonNullTypeNode):
        return recurse_type_annotation(
            field, graphql_type.type, schema, registry, optional=False
        )

    elif isinstance(graphql_type, ListTypeNode):
        return recurse_type_annotation(field, graphql_type.type, schema, registry)

    elif isinstance(graphql_type, NamedTypeNode):
        type = schema.get_type(graphql_type.name.value)
        assert type, graphql_type
        return recurse_type_annotation(field, type, schema, registry, optional=False)

    elif isinstance(graphql_type, GraphQLNonNull):
        return recurse_type_annotation(
            field, graphql_type.of_type, schema, registry, optional=False
        )

    elif isinstance(graphql_type, GraphQLList):
        return recurse_type_annotation(
            field, graphql_type.of_type, schema, registry, optional=False
        )

    elif isinstance(graphql_type, GraphQLScalarType):
        registry.register_scalar(graphql_type.name)

    elif isinstance(graphql_type, GraphQLEnumType):
        registry.register_enum(graphql_type.name)

    elif isinstance(graphql_type, GraphQLInputObjectType):
        # Only and only is we have not registered this input object type yet
        # we need to register it (otherwise we might get into a recursion loop)
        if not registry.is_input_registered(graphql_type.name):

            registry.register_input(graphql_type.name)

            # We need to get all input objects that this graphql input object type references

            for key, node in graphql_type.fields.items():
                recurse_type_annotation(node, node.type, schema, registry)

    else:
        raise Exception("Unknown Type", field, graphql_type)


def create_reference_registry_from_documents(
    schema: GraphQLSchema, document: DocumentNode
) -> ReferenceRegistry:
    """Finds all references of types of types that are used in the documents"""

    fragments: Dict[str, FragmentDefinitionNode] = {}
    operations: Dict[str, OperationDefinitionNode] = {}

    registry = ReferenceRegistry()

    for definition in document.definitions:
        if isinstance(definition, FragmentDefinitionNode):
            fragments[definition.name.value] = definition
        if isinstance(definition, OperationDefinitionNode):
            operations[definition.name.value] = definition

    for fragment in fragments.values():
        type = schema.get_type(fragment.type_condition.name.value)
        for selection in fragment.selection_set.selections:
            if isinstance(selection, FieldNode):
                # definition
                if selection.name.value == "__typename":
                    continue
                this_type = type.fields[selection.name.value]

                recurse_find_references(
                    selection,
                    this_type.type,
                    schema,
                    registry,
                )
            elif isinstance(selection, InlineFragmentNode):
                sub_type = schema.get_type(selection.type_condition.name.value)
                for sub_selection in selection.selection_set.selections:
                    if isinstance(sub_selection, FieldNode):
                        if sub_selection.name.value == "__typename":
                            continue
                        sub_this_type = sub_type.fields[sub_selection.name.value]
                        recurse_find_references(
                            sub_selection,
                            sub_this_type.type,
                            schema,
                            registry,
                        )

    for operation in operations.values():
        type = schema.get_root_type(operation.operation)

        for argument in operation.variable_definitions:
            recurse_type_annotation(argument, argument.type, schema, registry)

        for selection in operation.selection_set.selections:
            if isinstance(selection, FieldNode):
                # definition
                if selection.name.value == "__typename":
                    continue
                this_type = type.fields[selection.name.value]

                recurse_find_references(
                    selection,
                    this_type.type,
                    schema,
                    registry,
                )

    return registry



from dataclasses import dataclass
from graphql import (
    GraphQLInputObjectType,
    GraphQLInputType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLScalarType,
)
from pydantic_settings import SettingsConfigDict
from turms.plugins.base import Plugin, PluginConfig
import ast
from typing import Dict, List, Optional
from turms.config import GeneratorConfig
from graphql.utilities.build_client_schema import GraphQLSchema
from turms.plugins.base import Plugin
from pydantic import Field
from graphql.type.definition import (
    GraphQLEnumType,
)
from turms.referencer import create_reference_registry_from_documents
from turms.registry import ClassRegistry
from turms.utils import (
    generate_pydantic_config,
    get_additional_bases_for_type,
    parse_documents,
)
from turms.config import GraphQLTypes


class InputsPluginConfig(PluginConfig):
    model_config = SettingsConfigDict(
        extra="forbid", env_prefix="TURMS_PLUGINS_INPUTS_"
    )
    type: str = "turms.plugins.inputs.InputsPlugin"
    inputtype_bases: List[str] = ["pydantic.BaseModel"]
    allow_population_by_field_name: bool = True
    skip_underscore: bool = True
    skip_unreferenced: bool = True


def generate_input_annotation(
    type: GraphQLInputType,
    parent: str,
    config: GeneratorConfig,
    plugin_config: InputsPluginConfig,
    registry: ClassRegistry,
    is_optional=True,
):
    if isinstance(type, GraphQLScalarType):

        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_scalar(type.name),
            )

        return registry.reference_scalar(type.name)
        

    if isinstance(type, GraphQLInputObjectType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_inputtype(type.name, parent),
                ctx=ast.Load(),
            )
        return registry.reference_inputtype(type.name, parent)

    if isinstance(type, GraphQLEnumType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_enum(
                    type.name, parent, allow_forward=not config.force_plugin_order
                ),
                ctx=ast.Load(),
            )
        return registry.reference_enum(
            type.name, parent, allow_forward=not config.force_plugin_order
        )

    if isinstance(type, GraphQLNonNull):
        return generate_input_annotation(
            type.of_type, parent, config, plugin_config, registry, is_optional=False
        )

    if isinstance(type, GraphQLList):
        if (
            config.freeze.enabled
            and GraphQLTypes.INPUT in config.freeze.types
            and config.freeze.convert_list_to_tuple
        ):
            registry.register_import("typing.Tuple")

            def list_builder(x):
                return ast.Subscript(
                    value=ast.Name("Tuple", ctx=ast.Load()),
                    slice=ast.Tuple(elts=[x, ast.Constant(value=...)], ctx=ast.Load()),
                    ctx=ast.Load(),
                )

        else:
            registry.register_import("typing.List")

            def list_builder(x):
                return ast.Subscript(
                    value=ast.Name("List", ctx=ast.Load()), slice=x, ctx=ast.Load()
                )

        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=list_builder(
                    generate_input_annotation(
                        type.of_type,
                        parent,
                        config,
                        plugin_config,
                        registry,
                        is_optional=True,
                    )
                ),
                ctx=ast.Load(),
            )

        return list_builder(
            generate_input_annotation(
                type.of_type, parent, config, plugin_config, registry, is_optional=True
            )
        )

    raise NotImplementedError(f"Unknown input type {type}")



@dataclass
class Discriminator:
    discriminator: str
    value: str


def generate_input_type(
    name: str,
    union_type_descriminators: Dict[str, str],
    type: GraphQLInputType,
    config: GeneratorConfig,
    plugin_config: InputsPluginConfig,
    registry: ClassRegistry,
    key: str,
    discriminator: Optional[Discriminator] = None,
):
    
    additional_bases = get_additional_bases_for_type(type.name, config, registry)
    
    fields = (
        [ast.Expr(value=ast.Constant(value=type.description))]
        if type.description
        else []
    )

    if discriminator:
        fields.append(
            ast.AnnAssign(
                target=ast.Name(discriminator.discriminator, ctx=ast.Store()),
                annotation=ast.Subscript(value=ast.Name("Literal", ctx=ast.Load()), slice=ast.Constant(value=discriminator.value), ctx=ast.Load()),
                value=ast.Call(
                    func=ast.Name(id="Field", ctx=ast.Load()),
                    args=[],
                    keywords=[ast.keyword(arg="default", value=ast.Constant(value=discriminator.value))]
                ),
                simple=1,
            )
        )

    for value_key, value in type.fields.items():
        field_name = registry.generate_node_name(value_key)




        if field_name != value_key:
            registry.register_import("pydantic.Field")

            keywords = [
                ast.keyword(arg="alias", value=ast.Constant(value=value_key))
            ]
            if not isinstance(value.type, GraphQLNonNull):
                keywords.append(
                    ast.keyword(arg="default", value=ast.Constant(None))
                )


            assign = ast.AnnAssign(
                target=ast.Name(field_name, ctx=ast.Store()),
                annotation=generate_input_annotation(
                    value.type,
                    name,
                    config,
                    plugin_config,
                    registry,
                    is_optional=True,
                ),
                value=ast.Call(
                    func=ast.Name(id="Field", ctx=ast.Load()),
                    args=[],
                    keywords=keywords,
                ),
                simple=1,
            )

        else:
            assign = ast.AnnAssign(
                target=ast.Name(value_key, ctx=ast.Store()),
                annotation=generate_input_annotation(
                    value.type,
                    name,
                    config,
                    plugin_config,
                    registry,
                    is_optional=True,
                ),
                simple=1,
                value=(
                    ast.Constant(None)
                    if not isinstance(value.type, GraphQLNonNull)
                    else None
                ),
            )

        potential_comment = (
            value.description
            if not value.deprecation_reason
            else f"DEPRECATED: {value.description}"
        )

        if potential_comment:
            fields += [
                assign,
                ast.Expr(value=ast.Constant(value=potential_comment)),
            ]

        else:
            fields += [assign]

    

    return ast.ClassDef(
            name,
            bases=additional_bases
            + [
                ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                for base in plugin_config.inputtype_bases
            ],
            decorator_list=[],
            keywords=[],
            body=fields
            + generate_pydantic_config(
                GraphQLTypes.INPUT, config, registry, typename=key
            ),
        )






def generate_inputs(
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: InputsPluginConfig,
    registry: ClassRegistry,
):
    tree = []

    inputobjects_type = {
        key: value
        for key, value in client_schema.type_map.items()
        if isinstance(value, GraphQLInputObjectType)
    }

    if plugin_config.skip_unreferenced and config.documents:
        ref_registry = create_reference_registry_from_documents(
            client_schema, parse_documents(client_schema, config.documents)
        )
    else:
        ref_registry = None

    for base in plugin_config.inputtype_bases:
        registry.register_import(base)



    union_input_types = {}
    union_type_discriminators = {}

    for key, type in inputobjects_type.items():
        directives = type.ast_node.directives if type.ast_node else []
        for directive in directives:

            directive_name = directive.name.value
            if directive_name == "unionElementOf":

                union_type = None 
                discriminator = None
                key = None
                for arg in directive.arguments:
                    if arg.name.value == "union":
                        union_type = arg.value.value
                    if arg.name.value == "discriminator":
                        discriminator = arg.value.value
                    if arg.name.value == "key":
                        key = arg.value.value


                if union_type in ref_registry.inputs:
                    if union_type not in union_input_types:
                        union_input_types[union_type] = []
                    if union_type not in union_type_discriminators:
                        union_type_discriminators[union_type] = discriminator

                    assert union_type_discriminators[union_type] == discriminator, f"Discriminator mismatch for {union_type} expected {union_type_discriminators[union_type]} got {discriminator}"

                    name = registry.generate_inputtype(type.name)
                    union_input_types[union_type].append(name)
                    tree.append(generate_input_type(name, union_type_discriminators, type, config, plugin_config, registry, type.name, Discriminator(discriminator=discriminator, value=key)))

                
                




    for key, type in inputobjects_type.items():
        if ref_registry and key not in ref_registry.inputs:
            continue

        if plugin_config.skip_underscore and key.startswith("_"):  # pragma: no cover
            continue

        if type.name in union_input_types:
            registry.register_import("typing.Union")
            registry.register_import("typing.Annotated")
            registry.register_import("pydantic.Field")
            union_slice = ast.Tuple(
                elts=[
                    ast.Name(id=clsname, ctx=ast.Load())
                    for clsname in union_input_types[type.name]
                ],
                ctx=ast.Load(),
            )

            slice = ast.Tuple(
                elts=[
                    ast.Subscript(
                        value=ast.Name("Union", ctx=ast.Load()),
                        slice=union_slice,
                        ctx=ast.Load(),
                    ),
                    ast.Call(
                        func=ast.Name(id="Field", ctx=ast.Load()),
                        args=[],
                        keywords=[ast.keyword(arg="discriminator", value=ast.Constant(union_type_discriminators[type.name]))],
                    )
                ],
                ctx=ast.Load(),
            )




            tree.append(
                ast.Assign(
                    targets=[
                        ast.Name(
                            id=registry.generate_inputtype(type.name),
                            ctx=ast.Store(),
                        )
                    ],
                    value=ast.Subscript(
                        value=ast.Name("Annotated", ctx=ast.Load()),
                        slice=slice,
                        ctx=ast.Load(),
                    ),
                )
            )

            continue


        additional_bases = get_additional_bases_for_type(type.name, config, registry)
        name = registry.generate_inputtype(key)
        fields = (
            [ast.Expr(value=ast.Constant(value=type.description))]
            if type.description
            else []
        )

        for value_key, value in type.fields.items():
            field_name = registry.generate_node_name(value_key)

            if field_name != value_key:
                registry.register_import("pydantic.Field")
                keywords = [
                    ast.keyword(arg="alias", value=ast.Constant(value=value_key))
                ]
                if not isinstance(value.type, GraphQLNonNull):
                    keywords.append(
                        ast.keyword(arg="default", value=ast.Constant(None))
                    )

                assign = ast.AnnAssign(
                    target=ast.Name(field_name, ctx=ast.Store()),
                    annotation=generate_input_annotation(
                        value.type,
                        name,
                        config,
                        plugin_config,
                        registry,
                        is_optional=True,
                    ),
                    value=ast.Call(
                        func=ast.Name(id="Field", ctx=ast.Load()),
                        args=[],
                        keywords=keywords,
                    ),
                    simple=1,
                )

            else:
                assign = ast.AnnAssign(
                    target=ast.Name(value_key, ctx=ast.Store()),
                    annotation=generate_input_annotation(
                        value.type,
                        name,
                        config,
                        plugin_config,
                        registry,
                        is_optional=True,
                    ),
                    simple=1,
                    value=(
                        ast.Constant(None)
                        if not isinstance(value.type, GraphQLNonNull)
                        else None
                    ),
                )

            potential_comment = (
                value.description
                if not value.deprecation_reason
                else f"DEPRECATED: {value.description}"
            )

            if potential_comment:
                fields += [
                    assign,
                    ast.Expr(value=ast.Constant(value=potential_comment)),
                ]

            else:
                fields += [assign]

        

        tree.append(
            ast.ClassDef(
                name,
                bases=additional_bases
                + [
                    ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                    for base in plugin_config.inputtype_bases
                ],
                decorator_list=[],
                keywords=[],
                body=fields
                + generate_pydantic_config(
                    GraphQLTypes.INPUT, config, registry, typename=key
                ),
            )
        )


        



    return tree


class InputsPlugin(Plugin):
    """Generate pydantic Models for GraphQL inputs

    This plugin generates pydantic models for all GraphQL inputtypes in
    your schema. It will generate a model for each inputtype in your schema.

    """

    config: InputsPluginConfig = Field(default_factory=InputsPluginConfig)

    def generate_ast(
        self,
        client_schema: GraphQLSchema,
        config: GeneratorConfig,
        registry: ClassRegistry,
    ) -> List[ast.AST]:
        for base in self.config.inputtype_bases:
            registry.register_import(base)

        return generate_inputs(client_schema, config, self.config, registry)



import ast
import logging
from collections import defaultdict, deque
from typing import List, Optional

from graphql import (
    FieldNode,
    FragmentDefinitionNode,
    FragmentSpreadNode,
    GraphQLInterfaceType,
    GraphQLObjectType,
    GraphQLUnionType,
    InlineFragmentNode,
    SelectionSetNode,
    language,
)
from graphql.utilities.build_client_schema import GraphQLSchema
from graphql.utilities.type_info import get_field_def
from pydantic import Field
from pydantic_settings import SettingsConfigDict

from turms.config import GeneratorConfig, GraphQLTypes
from turms.plugins.base import Plugin, PluginConfig
from turms.recurse import type_field_node
from turms.registry import ClassRegistry
from turms.utils import (
    generate_generic_typename_field,
    generate_pydantic_config,
    generate_typename_field,
    get_additional_bases_for_type,
    get_interface_bases,
    non_typename_fields,
    parse_documents,
)

logger = logging.getLogger(__name__)


def find_fragment_dependencies_recursive(selection_set: SelectionSetNode, fragment_definitions, visited):
    """Recursively find all fragment dependencies within a selection set."""
    dependencies = set()
    if selection_set is None:
        return dependencies

    for selection in selection_set.selections:
        # If we encounter a fragment spread, add it to dependencies
        if isinstance(selection, FragmentSpreadNode):
            spread_name = selection.name.value
            dependencies.add(spread_name)
            # Recursively add dependencies of the spread fragment
            if spread_name in fragment_definitions and spread_name not in visited:
                visited.add(spread_name)  # Prevent cycles in recursion
                fragment = fragment_definitions[spread_name]
                dependencies.update(
                    find_fragment_dependencies_recursive(fragment.selection_set, fragment_definitions, visited)
                )
        # If it's a field with a nested selection set, dive deeper
        elif isinstance(selection, FieldNode) and selection.selection_set:
            dependencies.update(
                find_fragment_dependencies_recursive(selection.selection_set, fragment_definitions, visited)
            )


    return dependencies

def build_recursive_dependency_graph(document):
    """Build a dependency graph for fragments, accounting for deep nested fragment spreads."""
    fragment_definitions = {
        definition.name.value: definition for definition in document.definitions
        if isinstance(definition, FragmentDefinitionNode)
    }
    dependencies = defaultdict(set)
    
    # Populate the dependency graph with deeply nested fragment dependencies
    for fragment_name, fragment in fragment_definitions.items():
        visited = set()  # Track visited fragments to avoid cyclic dependencies
        dependencies[fragment_name] = find_fragment_dependencies_recursive(fragment.selection_set, fragment_definitions, visited)
    
    return dependencies


def topological_sort(dependency_graph):
    """Perform a topological sort on fragments based on recursive dependencies."""
    sorted_fragments = []
    no_dependency_fragments = deque([frag for frag, deps in dependency_graph.items() if not deps])
    resolved = set(no_dependency_fragments)
    
    while no_dependency_fragments:
        fragment = no_dependency_fragments.popleft()
        sorted_fragments.append(fragment)
        
        # Remove this fragment from other fragments' dependencies
        for frag, deps in dependency_graph.items():
            if fragment in deps:
                deps.remove(fragment)
                if not deps and frag not in resolved:
                    no_dependency_fragments.append(frag)
                    resolved.add(frag)
    
    # Add any remaining fragments that may have been missed if they were independent
    sorted_fragments.extend(frag for frag in dependency_graph if frag not in sorted_fragments)
    
    return sorted_fragments


class FragmentsPluginConfig(PluginConfig):
    model_config = SettingsConfigDict(env_prefix="TURMS_PLUGINS_FRAGMENTS_")
    type: str = "turms.plugins.fragments.FragmentsPlugin"
    fragment_bases: List[str] = []
    fragments_glob: Optional[str] = None
    add_documentation: bool = True


def get_fragment_bases(
    config: GeneratorConfig,
    pluginConfig: FragmentsPluginConfig,
    registry: ClassRegistry,
):
    if pluginConfig.fragment_bases:
        for base in pluginConfig.fragment_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in pluginConfig.fragment_bases
        ]

    else:
        for base in config.object_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in config.object_bases
        ]


def get_implementing_types(type: GraphQLInterfaceType, client_schema: GraphQLSchema):
    implementing_types = []
    for type in client_schema.get_implementing_types(type):
        implementing_types.append(type)
        implementing_types += get_implementing_types(type, client_schema)
    return implementing_types



def generate_fragment(
    f: FragmentDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: FragmentsPluginConfig,
    registry: ClassRegistry,
):
    tree = []
    fields = []
    type = client_schema.get_type(f.type_condition.name.value)
    name = registry.generate_fragment(
        f.name.value, isinstance(type, GraphQLInterfaceType)
    )

    registry.register_fragment_document(
        f.name.value, language.print_ast(f)
    )  # TODO: Check if typename is being referenced? so that we can check between the elements of the interface

    registry.register_fragment_type(f.name.value, type)




    if isinstance(type, GraphQLInterfaceType):

        implementing_types = client_schema.get_implementations(type)
    
        mother_class_fields = []
        base_fragment_name =  registry.style_fragment_class(f.name.value)
        additional_bases = get_additional_bases_for_type(type.name, config, registry)

        if type.description and plugin_config.add_documentation:
            mother_class_fields.append(
                ast.Expr(value=ast.Constant(value=type.description))
            )

        sub_nodes = non_typename_fields(f)

        mother_class_name = base_fragment_name + "Base"


        implementing_class_base_classes = {
        }


        inline_fragment_fields = {}



        for sub_node in sub_nodes:

            if isinstance(sub_node, FragmentSpreadNode):
                # Spread nodes are like inheritance?
                try:
                    # We are dealing with a fragment that is an interface
                    implementation_map = registry.get_interface_fragment_implementations(sub_node.name.value)
                    for k, v in implementation_map.items():
                        implementing_class_base_classes.setdefault(k, []).append(v)

                except KeyError:
                    x = registry.get_fragment_type(sub_node.name.value)
                    implementing_class_base_classes.setdefault(x, []).append(registry.inherit_fragment(sub_node.name.value))



            if isinstance(sub_node, FieldNode):

                field_type = type.fields[sub_node.name.value]
                mother_class_fields += type_field_node(
                    sub_node,
                    base_fragment_name,
                    field_type,
                    client_schema,
                    config,
                    tree,
                    registry,
                )

            if isinstance(sub_node, InlineFragmentNode):
                on_type_name = sub_node.type_condition.name.value

                inline_fragment_fields.setdefault(on_type_name, []).append(
                    generate_typename_field(
                        sub_node.type_condition.name.value, registry, config
                    )
                )


        mother_class = ast.ClassDef(
            mother_class_name,
            bases=additional_bases + get_interface_bases(config, registry) ,  # Todo: fill with base
            decorator_list=[],
            keywords=[],
            body=mother_class_fields if mother_class_fields else [ast.Pass()],
        )

        catch_class_name = f"{base_fragment_name}Catch"

        catch_class = ast.ClassDef(
            catch_class_name,
            bases=[ast.Name(id=mother_class_name, ctx=ast.Load())],  # Todo: fill with base
            decorator_list=[],
            keywords=[],
            body=[generate_generic_typename_field(registry, config)] + mother_class_fields,
        )



        tree.append(mother_class)
        tree.append(catch_class)



        implementaionMap = {}

        for i in implementing_types.objects:

            class_name = f"{base_fragment_name}{i.name}"


            ast_base_nodes = [ast.Name(id=x, ctx=ast.Load()) for x in implementing_class_base_classes.get(i, [])]
            implementaionMap[i.name] = class_name

            inline_fields = inline_fragment_fields.get(i, [])

            implementing_class = ast.ClassDef(
                    class_name,
                    bases=ast_base_nodes + [ast.Name(id=mother_class_name, ctx=ast.Load())] + get_interface_bases(config, registry),  # Todo: fill with base
                    decorator_list=[],
                    keywords=[],
                    body=[generate_typename_field(i.name, registry, config)] + inline_fields,
            )

            tree.append(implementing_class)


        registry.register_interface_fragment_implementations(f.name.value, implementaionMap)


        return tree


    elif isinstance(type, GraphQLObjectType):
        additional_bases = get_additional_bases_for_type(
            f.type_condition.name.value, config, registry
        )

        if type.description and plugin_config.add_documentation:
            fields.append(
                ast.Expr(value=ast.Constant(value=type.description))
            )

        fields += [generate_typename_field(type.name, registry, config)]

        for field in f.selection_set.selections:

            if field.name.value == "__typename":
                continue

            if isinstance(field, FragmentDefinitionNode):  # pragma: no cover

                continue

            if isinstance(field, FragmentSpreadNode):
                try:
                    implementationMap = registry.get_interface_fragment_implementations(field.name.value)
                    if type.name in implementationMap:
                        additional_bases = [
                            ast.Name(
                                id=implementationMap[type.name],
                                ctx=ast.Load(),
                            )
                        ] + additional_bases
                    else:
                        raise Exception(f"Could not find implementation for {type.name} in {implementationMap}")
                except KeyError:
                    additional_bases = [
                        ast.Name(
                            id=registry.inherit_fragment(field.name.value),
                            ctx=ast.Load(),
                        )
                    ] + additional_bases  # needs to be prepended (MRO)
                continue

            field_definition = get_field_def(client_schema, type, field)
            assert field_definition, "Couldn't find field definition"

            fields += type_field_node(
                field,
                name,
                field_definition,
                client_schema,
                config,
                tree,
                registry,
            )

        tree.append(
            ast.ClassDef(
                name,
                bases=additional_bases
                + get_fragment_bases(config, plugin_config, registry),
                decorator_list=[],
                keywords=[],
                body=fields
                + generate_pydantic_config(GraphQLTypes.FRAGMENT, config, registry),
            )
        )
        return tree

    elif isinstance(type, GraphQLUnionType):
        base_fragment_name = registry.style_fragment_class(f.name.value)
        additional_bases = get_additional_bases_for_type(type.name, config, registry)

        sub_nodes = non_typename_fields(f)

        mother_class_name = base_fragment_name  # + "Base"

        member_class_base_classes = {}

        inline_fragment_fields = {}

        for sub_node in sub_nodes:

            if isinstance(sub_node, FragmentSpreadNode):
                # Spread nodes are like inheritance
                # We are dealing with a fragment that is a union
                implementation_map = registry.get_union_fragment_members(
                    sub_node.name.value
                )
                for k, v in implementation_map.items():
                    member_class_base_classes.setdefault(k, []).append(v)

            elif isinstance(sub_node, FieldNode):
                if sub_node.name.value == "__typename":
                    continue
                raise AssertionError("Union types should not have fields")

            elif isinstance(sub_node, InlineFragmentNode):
                on_type_name = sub_node.type_condition.name.value

                fields = []
                for field in sub_node.selection_set.selections:

                    if field.name.value == "__typename":
                        continue

                    if isinstance(field, FragmentSpreadNode):
                        member_class_base_classes.setdefault(on_type_name, []).append(
                            field.name.value
                        )
                        continue

                    field_type = client_schema.get_type(on_type_name)

                    field_definition = get_field_def(
                        client_schema, field_type, field
                    )
                    assert (
                        field_definition
                    ), f"Couldn't find field definition for {on_type_name}.{field.name.value}"

                    fields += type_field_node(
                        field,
                        name,
                        field_definition,
                        client_schema,
                        config,
                        tree,
                        registry,
                    )

                inline_fragment_fields.setdefault(on_type_name, []).extend(fields)
            else:
                raise AssertionError(f"Unknown node type: {type(sub_node)}")

        implementationMap = {}

        for i in type.types:

            class_name = f"{base_fragment_name}{i.name}"

            ast_base_nodes = [
                ast.Name(id=x, ctx=ast.Load())
                for x in member_class_base_classes.get(i.name, [])
            ]
            implementationMap[i.name] = class_name

            inline_fields = inline_fragment_fields.get(i.name, [])

            implementing_class = ast.ClassDef(
                class_name,
                bases=ast_base_nodes
                + get_fragment_bases(config, plugin_config, registry),
                decorator_list=[],
                keywords=[],
                body=[generate_typename_field(i.name, registry, config)]
                + inline_fields,
            )

            tree.append(implementing_class)

        registry.register_union_fragment_members(
            f.name.value, implementationMap
        )

        registry.register_import("typing.Union")
        mother_class = ast.Assign(
            targets=[ast.Name(id=base_fragment_name, ctx=ast.Load())],
            value=ast.Subscript(
                value=ast.Name(id="Union", ctx=ast.Load()),
                slice=ast.Tuple(
                    elts=[
                        ast.Name(id=f"{base_fragment_name}{i.name}", ctx=ast.Load())
                        for i in type.types
                    ],
                    ctx=ast.Load(),
                ),
            ),
            simple=1,
        )

        tree.append(mother_class)

        if type.description and plugin_config.add_documentation:
            tree.append(
                ast.Expr(value=ast.Constant(value=type.description))
            )

        return tree


def reorder_definitions(document, sorted_fragments):
    """Reorder document definitions to place fragments in dependency order."""
    fragment_definitions = {defn.name.value: defn for defn in document.definitions if isinstance(defn, FragmentDefinitionNode)}

    # Order fragments according to the topologically sorted order
    ordered_fragments = [fragment_definitions[name] for name in sorted_fragments if name in fragment_definitions]

    # Combine operations and ordered fragments
    return ordered_fragments


class FragmentsPlugin(Plugin):
    """Plugin for generating fragments from
    documents

    The fragments plugin will generate classes for each fragment. It loads the documents,
    scans for fragments and generates the classes.

    If encountering a fragment on an interface it will generate a BASE class for that interface
    and then generate a class for each type referenced in the fragment. They will all inherit
    from the base class. The true type will be determined at runtime as all of the potential subtypes
    will be in the same union.

    """

    config: FragmentsPluginConfig = Field(default_factory=FragmentsPluginConfig)

    def generate_ast(
        self,
        client_schema: GraphQLSchema,
        config: GeneratorConfig,
        registry: ClassRegistry,
    ) -> List[ast.AST]:

        plugin_tree = []

        documents = parse_documents(
            client_schema, self.config.fragments_glob or config.documents
        )

        # Find dependencies and sort fragments topologically
        fragment_dependencies = build_recursive_dependency_graph(documents)
       
        sorted_fragments = topological_sort(fragment_dependencies)
        


        ordered_fragments = reorder_definitions(documents, sorted_fragments)

        
        

        for fragment in ordered_fragments:
            plugin_tree += generate_fragment(
                fragment, client_schema, config, self.config, registry
            )

        return plugin_tree



from pydantic import Field
from turms.stylers.base import BaseStyler, StylerConfig


class AppenderStylerConfig(StylerConfig):
    type: str = "turms.stylers.appender.AppenderStyler"
    append_fragment: str = "Fragment"
    append_query: str = "Query"
    append_mutation: str = "Mutation"
    append_subscription: str = "Subscription"
    append_enum: str = ""
    append_input: str = ""


class AppenderStyler(BaseStyler):
    config: AppenderStylerConfig = Field(default_factory=AppenderStylerConfig)

    def style_fragment_name(self, typename):
        return f"{typename}{self.config.append_fragment}"

    def style_query_name(self, typename):
        return f"{typename}{self.config.append_query}"

    def style_subscription_name(self, typename):
        return f"{typename}{self.config.append_subscription}"

    def style_mutation_name(self, typename):
        return f"{typename}{self.config.append_mutation}"

    def style_input_name(self, typename):
        return f"{typename}{self.config.append_input}"

    def style_enum_name(self, typename):
        return f"{typename}{self.config.append_enum}"



import json
from importlib import import_module
from typing import Any, Dict, Optional, Tuple
import glob
from pydantic import AnyHttpUrl
from turms.errors import GenerationError

from graphql import (
    get_introspection_query,
)

IntrospectionResult = Dict[str, Any]
DSLString = str


def import_class(module_path, class_name):
    """Import a module from a module_path and return the class"""
    module = import_module(module_path)
    return getattr(module, class_name)


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed. Simliar to
    djangos import_string, but without the cache.
    """

    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as err:
        raise ImportError(f"{dotted_path} doesn't look like a module path") from err

    try:
        return import_class(module_path, class_name)
    except AttributeError as err:
        raise ImportError(
            f"{module_path} does not define a {class_name} attribute/class"
        ) from err


def load_introspection_from_url(
    url: AnyHttpUrl, headers: Optional[Dict[str, str]] = None
) -> IntrospectionResult:
    """Introspect a GraphQL schema using introspection query

    Args:
        schema_url (str): The Schema url
        bearer_token (str, optional): A Bearer token. Defaults to None.

    Raises:
        GenerationError: An error occurred while generating the schema.

    Returns:
        dict: The introspection query response.
    """
    try:  # pragma: no cover
        import requests  # pragma: no cover
    except ImportError:  # pragma: no cover
        raise GenerationError(
            "The requests library is required to introspect a schema from a url"
        )  # pragma: no cover

    jdata = json.dumps({"query": get_introspection_query()}).encode("utf-8")
    default_headers = {"Content-Type": "application/json", "Accept": "application/json"}
    if headers:
        default_headers.update(headers)
    try:
        req = requests.post(url, data=jdata, headers=default_headers)
        x = req.json()
    except Exception:
        raise GenerationError(f"Failed to fetch schema from {url}")
    if "errors" in x:  # pragma: no cover
        raise GenerationError(
            f"Failed to fetch schema from {url} Graphql error: {x['errors']}"
        )

    if "data" not in x:
        raise GenerationError(
            f"Failed to fetch schema from {url}. Did not receive data attripute: {x}"
        )

    return x["data"]


def load_dsl_from_url(url: AnyHttpUrl, headers: Dict[str, str] = None) -> DSLString:
    try:  # pragma: no cover
        import requests  # pragma: no cover
    except ImportError:  # pragma: no cover
        raise GenerationError(
            "The requests library is required to introspect a schema from a url"
        )  # pragma: no cover

    default_headers = {}
    if headers:
        default_headers.update(headers)
    try:
        req = requests.get(url, headers=default_headers)
        assert req.status_code == 200, "Incorrect status code"
        assert req.content, "No content"
        x = req.content.decode()
    except Exception as e:
        raise GenerationError(f"Failed to fetch schema from {url}") from e
    return x


def load_dsl_from_file(file_path: str) -> DSLString:
    """Load a GraphQL DSL file and return its content as a string"""
    with open(file_path, "rb") as f:
        return f.read().decode("utf-8-sig")


def load_introspection_from_file(file_path: str) -> IntrospectionResult:
    """Load a GraphQL introspection file and return its content as a string"""
    with open(file_path, "rb") as f:
        return json.loads(f.read().decode("utf-8-sig"))


ParseResult = Tuple[Optional[DSLString], Optional[IntrospectionResult]]


def load_dsl_from_glob(
    glob_string: str,
    allow_introspection: bool = True,
) -> DSLString:
    """Build a GraphQL schema from a glob string"""
    schema_glob = glob.glob(glob_string, recursive=True)
    if len(schema_glob) == 0:
        raise GenerationError(f"No files found for glob string {glob_string}")

    dsl_string = ""
    for file in schema_glob:
        dsl_string += load_dsl_from_file(file)

    return dsl_string


def load_introspection_from_glob(
    glob_string: str,
):
    schema_glob = glob.glob(glob_string, recursive=True)
    if len(schema_glob) == 0:
        raise GenerationError(f"No files found for glob string {glob_string}")

    if len(schema_glob) > 1:
        raise GenerationError(
            f"More than one file found for glob string {glob_string}. Introspection can only be loaded from one file."
        )

    return load_introspection_from_file(schema_glob[0])



from enum import Enum
import os
from typing import Dict
from turms.config import GraphQLProject
from turms.run import generate, write_code_to_file, write_project, write_schema_to_file
from rich import get_console
from rich.panel import Panel
from rich.live import Live
from rich.tree import Tree
from rich.console import Group
import rich_click as click
from turms.run import (
    scan_folder_for_single_config,
    load_projects_from_configpath,
    build_schema_from_schema_type,
)
from .watch import stream_changes
from graphql import print_schema
from functools import wraps

click.rich_click.USE_RICH_MARKUP = True

directory = os.getcwd()


class TurmsOptions(str, Enum):
    GEN = "gen"
    INIT = "init"
    DOWNLOAD = "download"
    WATCH = "watch"


logo = """
   __                             
  / /___  ___________ ___  _____
 / __/ / / / ___/ __ `__ \/ ___/
/ /_/ /_/ / /  / / / / / (__  ) 
\__/\__,_/_/  /_/ /_/ /_/____/  
"""

welcome = """
Welcome to Turms! Turms is a GraphQL code generator that generates code from your GraphQL schema and documents. For more information, visit
https://github.com/jhnnsrs/turms
"""


default_settings = """
projects:
  default:
    schema: https://countries.trevorblades.com/
    documents: graphql/**.graphql
    extensions:
      turms:
        out_dir: examples/api
        plugins:
          - type: turms.plugins.enums.EnumsPlugin
          - type: turms.plugins.inputs.InputsPlugin
          - type: turms.plugins.fragments.FragmentsPlugin
          - type: turms.plugins.operations.OperationsPlugin
          - type: turms.plugins.funcs.FuncsPlugin
        processors:
          - type: turms.processors.black.BlackProcessor
        scalar_definitions:
          uuid: str
"""


def generate_projects(projects: Dict[str, GraphQLProject], title="Turms"):
    generation_message = f"Generating the {'.'.join(projects.keys())} projects. This may take a while...\n"

    tree = Tree("Generating projects", style="bold green")
    panel_group = Group(generation_message, tree)

    panel = Panel(
        panel_group,
        title=title,
        title_align="left",
        border_style="green",
        padding=(1, 1),
    )

    raised_exceptions = []

    with Live(panel, screen=False) as live:
        for key, project in projects.items():
            project_tree = Tree(f"{key}", style="not bold white")
            tree.add(project_tree)
            live.update(panel)

            def log(message, level):
                if level == "WARN":
                    project_tree.add(Tree(message, style="yellow"))

            try:
                generated_code, schema = generate(project, log=log)

                project_tree.label = f"{key} ✔️"
                live.update(panel)

                write_code_to_file(
                    generated_code,
                    project.extensions.turms.out_dir,
                    project.extensions.turms.generated_name,
                )

                if project.extensions.turms.dump_schema:
                    write_schema_to_file(schema, project.extensions.turms.out_dir, project.extensions.turms.schema_name)

                if project.extensions.turms.dump_configuration:
                    write_project(project, project.extensions.turms.out_dir, project.extensions.turms.configuration_name)

            except Exception as e:
                project_tree.style = "red"
                project_tree.label = f"{key} 💥"
                project_tree.add(Tree(str(e), style="red"))
                if project.extensions.turms.exit_on_error:
                    raised_exceptions.append(e)
                live.update(panel)

    if raised_exceptions:
        # print traceback of first exception
        for e in raised_exceptions:
            try:
                raise e
            except Exception:
                get_console().print_exception()

        raise click.ClickException(
            "One or more projects failed to generate. First error"
        ) from raised_exceptions[0]


def with_projects(func):
    @click.argument("project", default=None, required=False)
    @click.option("--config", default=None)
    @wraps(func)
    def wrapper(*args, config=None, project=None, **kwargs):
        try:
            app_directory = os.getcwd()
            config = config or scan_folder_for_single_config(app_directory)
            if not config:
                raise Exception(
                    f"No config file found. Please run `turms init` in {app_directory} to create a default config file or specify a config file with the --config flag"
                )

            try:
                projects = load_projects_from_configpath(config)
            except Exception as e:
                get_console().print_exception()
                raise e

            if project:
                projects = {
                    key: value for key, value in projects.items() if key == project
                }
                if not projects:
                    raise Exception(f"Project {project} not found in {config}")

        except Exception as e:
            raise click.ClickException(str(e)) from e

        return func(*args, projects=projects, **kwargs)

    return wrapper


def watch_projects(projects, title="Turms"):  # pragma: no cover
    if len(projects) > 1:
        raise click.ClickException(
            "Watching multiple projects is not supported. Please specify a single project!"
        )

    project = list(projects.values())[0]

    generation_message = f"Watching the {'.'.join(projects.keys())} project. Changes will be automatically generated and added to [b]{project.extensions.turms.out_dir}/{project.extensions.turms.generated_name}[/b] when you save a file in the project's documents folder."

    tree = Panel("Watching....", style="bold green")
    panel_group = Group(generation_message, tree)

    panel = Panel(
        panel_group,
        title=title,
        title_align="left",
        border_style="green",
        padding=(1, 1),
    )

    with Live(panel, screen=False) as live:
        tree.renderable = f"Watching {project.documents}..."
        live.update(panel)

        for event in stream_changes(project.documents):
            live.update(panel)

            try:
                tree.renderable = "Generating..."
                tree.border_style = "blue"
                tree.style = "blue"
                live.update(panel)

                generated_code, schema = generate(
                    project,
                )

                write_code_to_file(
                    generated_code,
                    project.extensions.turms.out_dir,
                    project.extensions.turms.generated_name,
                )

                tree.renderable = "Generation Successfull"
                tree.border_style = "green"
                tree.style = "green"
                live.update(panel)

            except Exception as e:
                tree.renderable = f"[red] {str(e)} [/],"
                tree.border_style = "not bold red"
                tree.style = "not bold red"
                live.update(panel)
                continue


@click.group()
@click.pass_context
def cli(ctx):
    """Welcome to turms!

    Welcome to Turms! Turms is a GraphQL code generator that generates code from your GraphQL schema and documents.

    For more information, visit [link=https://github.com/jhnnsrs/turms] https://github.com/jhnnsrs/turms [/link]
    """


@cli.command()
@click.option("--config", default="graphql.config.yaml", help="The config file to use")
def init(config):
    """Initialize a new graphql project"""
    welcome_panel = Panel(logo + welcome, title="turms", border_style="bold green")

    get_console().print(welcome_panel)

    app_directory = os.getcwd()
    if os.path.exists(os.path.join(app_directory, "graphql.config.yaml")):
        if not click.confirm(
            f"Config file already exists in {app_directory}. Do you want to overwrite it?",
            default=False,
        ):
            get_console().print("Aborting")
            return

    get_console().print(f"Creating demo graphql.config.yaml in {app_directory}")
    with open(os.path.join(app_directory, "graphql.config.yaml"), "w") as f:
        f.write(default_settings)


@cli.command()
@with_projects
def gen(projects):
    """Generate the graphql project"""
    generate_projects(projects)


@cli.command()
@with_projects
def watch(projects):  # pragma: no cover
    """Watch the graphql project"""
    watch_projects(projects)


@cli.command()
@with_projects
@click.option(
    "--out",
    default=".schema.graphql",
    help="The output file extension (will be appended to the project name)",
)
@click.option(
    "--dir",
    default=None,
    help="The output directory for the schema files (will default to the current working directory)",
)
def download(projects, out, dir):
    """Download the graphql projects schema as a sdl file"""

    try:
        app_directory = dir or os.getcwd()
        for key, project in projects.items():
            filename = f"{key}{out}"
            get_console().print(
                f"Downloading schema for project {key} to {app_directory}/{filename}"
            )
            schema = build_schema_from_schema_type(
                project.schema_url, allow_introspection=True
            )
            with open(os.path.join(app_directory, filename), "w") as f:
                f.write(print_schema(schema))
    except Exception as e:
        raise click.ClickException(str(e)) from e


if __name__ == "__main__":
    cli()
